""" Generators for basis evaluations """

from dune.perftool.generation import (cached,
                                      domain,
                                      iname,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.quadrature import (name_quadraturepoint,
                                             quadrature_iname,
                                             )
from dune.perftool.pdelab.geometry import (name_jacobian_inverse_transposed,
                                           )


@symbol
def lfs_iname_bound(number):
    return "arg{}_n".format(number)


@iname
def _lfs_iname(element, argcount):
    ainame = "arg{}".format(chr(ord("i") + argcount))
    domain(ainame, lfs_iname_bound(argcount))
    return ainame


def lfs_iname(element, argcount=0):
    return _lfs_iname(element, argcount)


@symbol
def name_localfunctionspace(expr):
    return 'lfsu'


@cached
def evaluate_basis(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=('arg0_n',))
    lfs = name_localfunctionspace(element)
    qp = name_quadraturepoint()
    instruction(inames=(quadrature_iname(),
                        ),
                code='{}.finiteElement().localBasis().evaluateFunction({}, {});'.format(lfs,
                                                                                        qp,
                                                                                        name,
                                                                                        ),
                assignees=name,
                )


@symbol
def name_basis(element):
    evaluate_basis(element, "phi")
    return "phi"


@cached
def evaluate_reference_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=('arg0_n', 2))
    lfs = name_localfunctionspace(element)
    qp = name_quadraturepoint()
    instruction(inames=(quadrature_iname(),
                        ),
                code='{}.finiteElement().localBasis().evaluateJacobian({}, {});'.format(lfs,
                                                                                        qp,
                                                                                        name,
                                                                                        ),
                assignees=name,
                )


@symbol
def name_reference_gradient(element):
    evaluate_reference_gradient(element, "js")
    return "js"


@cached
def evaluate_basis_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=('arg0_n', 2))
    jac = name_jacobian_inverse_transposed()
    index = lfs_iname(element)
    reference_gradients = name_reference_gradient(element)
    instruction(inames=(index,
                        quadrature_iname(),
                        ),
                code='{}.mv({}[{}][0], {});'.format(jac,
                                                    reference_gradients,
                                                    index,
                                                    name,
                                                    ),
                assignees=name,
                )


@symbol
def name_basis_gradient(element):
    evaluate_basis_gradient(element, "gradphi")
    return "gradphi"


@cached
def evaluate_trialfunction(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=())
    lfs = name_localfunctionspace(element)
    index = lfs_iname(element)
    basis = name_basis()
    instruction(inames=(quadrature_iname(),
                        index,
                        ),
                code='{} += x({}, {})*{}[{}];'.format(name,
                                                      lfs,
                                                      index,
                                                      basis,
                                                      index
                                                      ),
                assignees=name,
                )


@cached
def evaluate_trialfunction_gradient(element, name):
    # TODO this is of course not yet correct
    temporary_variable(name, shape=(2,))
    lfs = name_localfunctionspace(element)
    index = lfs_iname(element)
    basis = name_basis_gradient(element)
    instruction(inames=(quadrature_iname(),
                        index,
                        ),
                code='{}.axpy(x({}, {}), {});'.format(name,
                                                      lfs,
                                                      index,
                                                      basis
                                                      ),
                assignees=name,
                )
